/*
 * Administrators Control Module Version 1.02 For Kernels 2.2.x
 * By xfer (xfer@oko.lo2.szczecin.pl)
 * At Sat Feb 19 18:13:14 CET 2000
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#define __KERNEL__
#define MODULE
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/module.h>
#include <linux/unistd.h>
#include <linux/stat.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/utime.h>
#include <linux/shm.h>
#include <linux/ipc.h>
#include <sys/syscall.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/ptrace.h>
#define VERSION "1.02"
#define RESTRICTED_GID 100
#define RESTRICTED_TEXT_FILE "/etc/group"
#define RESTRICTED_BIN_FILE "/bin/ping"
extern void *sys_call_table[];
int (*original_setuid) (uid_t uid);
int (*original_setgid) (gid_t gid);
int (*original_setreuid) (uid_t ruid, uid_t euid);
int (*original_setregid) (gid_t rgid, gid_t egid);
int (*original_ipc) (unsigned int call, int first, int second, int third, void *ptr, long fifth);
int (*original_execve) (struct pt_regs regs);
int (*original_open) (const char *pathname, int flags, mode_t mode);
int (*getuid) (void);
int (*geteuid) (void);
int (*getgid) (void);
int (*getegid) (void);
int (*stat) (const char *file_name, struct stat * buf);
int
acm_setuid (uid_t uid)
{
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      printk (KERN_INFO "User [%d] was trying to setuid [%d]\n", getuid (), uid);
    }
  return (*original_setuid) (uid);
}
int
acm_setgid (gid_t gid)
{
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      printk (KERN_INFO "User [%d] was trying to setgid [%d]\n", getuid (), gid);
    }
  return (*original_setgid) (gid);
}
int
acm_setreuid (uid_t ruid, uid_t euid)
{
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      printk (KERN_INFO "User [%d] was trying to setreuid [%d/%d]\n", getuid (), ruid, euid);
    }
  return (*original_setreuid) (ruid, euid);
}
int
acm_setregid (gid_t rgid, gid_t egid)
{
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      printk (KERN_INFO "User [%d] was trying to setregid [%d/%d]\n", getuid (), rgid, egid);
    }
  return (*original_setregid) (rgid, egid);
}
int
acm_ipc (unsigned int call, int first, int second, int third, void *ptr, long fifth)
{
  if (current->euid)
    return -EPERM;
  return original_ipc (call, first, second, third, ptr, fifth);
}
int
acm_open (const char *pathname, int flags, mode_t mode)
{
  struct stat acm_file, original_file;
  mm_segment_t f_sys;
  int x;
  lock_kernel ();
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      f_sys = get_fs ();
      set_fs (KERNEL_DS);
      if (stat (RESTRICTED_TEXT_FILE, &acm_file))
	printk (KERN_INFO "acm: open(): stat(): %s\n", RESTRICTED_TEXT_FILE);
      if (stat (pathname, &original_file))
	{
	  return original_open (pathname, flags, mode);
	}
      if (acm_file.st_ino == original_file.st_ino && acm_file.st_dev == original_file.st_dev)
	{
	  printk (KERN_INFO "User [%d] was trying to open forbidden file [%s]\n", getuid (), RESTRICTED_TEXT_FILE);
	  return -EACCES;
	}
      set_fs (f_sys);
      return original_open (pathname, flags, mode);
    }
  unlock_kernel ();
  return original_open (pathname, flags, mode);
}
int
acm_execve (struct pt_regs regs)
{
  int ret;
  char *filename, *tmp, *ar = NULL, **argv = NULL;
  struct stat acm_file, original_file;
  mm_segment_t f_sys;
  lock_kernel ();
  filename = getname ((char *) regs.ebx);
  if (getgid () == RESTRICTED_GID && getegid () == RESTRICTED_GID && getuid () != 0 && geteuid () != 0)
    {
      f_sys = get_fs ();
      set_fs (KERNEL_DS);
      if (stat (RESTRICTED_BIN_FILE, &acm_file))
	printk (KERN_INFO "acm: execve(): stat(): %s\n", RESTRICTED_BIN_FILE);
      if (stat (filename, &original_file))
	printk (KERN_INFO "acm: execve(): stat(): %s\n", filename);
      if (acm_file.st_ino == original_file.st_ino && acm_file.st_dev == original_file.st_dev)
	{
	  printk (KERN_INFO "User [%d] was trying to execute forbidden file [%s]\n", getuid (), RESTRICTED_BIN_FILE);
	  return -EACCES;
	}
      if (original_file.st_mode & S_ISUID)
	{
	  printk (KERN_INFO "User [%d] executed SUID program [ %s ", getuid (), filename);
	  argv = (char **) regs.ecx;
	  get_user (ar, ++argv);
	  while (*argv && ar != NULL)
	    {
	      if (IS_ERR (tmp = getname (*argv)))
		break;
	      printk ("%s ", tmp);
	      get_user (ar, ++argv);
	      putname (tmp);
	    }
	  printk ("]\n");
	}
      if (original_file.st_mode & S_ISGID)
	{
	  printk (KERN_INFO "User [%d] executed SGID program [ %s ", getuid (), filename);
	  argv = (char **) regs.ecx;
	  get_user (ar, ++argv);
	  while (*argv && ar != NULL)
	    {
	      if (IS_ERR (tmp = getname (*argv)))
		break;
	      printk ("%s ", tmp);
	      get_user (ar, ++argv);
	      putname (tmp);
	    }
	  printk ("]\n");
	}
      set_fs (f_sys);
    }
  ret = do_execve (filename, (char **) regs.ecx, (char **) regs.edx, &regs);
  putname (filename);
  unlock_kernel ();
  return ret;
}
int
init_module (void)
{
  stat = sys_call_table[__NR_stat];
  getuid = sys_call_table[__NR_getuid];
  geteuid = sys_call_table[__NR_geteuid];
  getgid = sys_call_table[__NR_getgid];
  getegid = sys_call_table[__NR_getegid];
  original_setuid = sys_call_table[__NR_setuid];
  sys_call_table[__NR_setuid] = acm_setuid;
  original_setgid = sys_call_table[__NR_setgid];
  sys_call_table[__NR_setgid] = acm_setgid;
  original_setreuid = sys_call_table[__NR_setreuid];
  sys_call_table[__NR_setreuid] = acm_setreuid;
  original_setregid = sys_call_table[__NR_setregid];
  sys_call_table[__NR_setregid] = acm_setregid;
  original_ipc = sys_call_table[__NR_ipc];
  sys_call_table[__NR_ipc] = acm_ipc;
  original_open = sys_call_table[__NR_open];
  sys_call_table[__NR_open] = acm_open;
  original_execve = sys_call_table[__NR_execve];
  sys_call_table[__NR_execve] = acm_execve;
  printk ("Administrators Control Module Version %s Installed\n", VERSION);
  return 0;
}
void
cleanup_module (void)
{
  sys_call_table[__NR_setuid] = original_setuid;
  sys_call_table[__NR_setgid] = original_setgid;
  sys_call_table[__NR_setreuid] = original_setreuid;
  sys_call_table[__NR_setregid] = original_setregid;
  sys_call_table[__NR_ipc] = original_ipc;
  sys_call_table[__NR_open] = original_open;
  sys_call_table[__NR_execve] = original_execve;
  printk ("Administrators Control Module Version %s Uninstalled\n", VERSION);
}
